from __future__ import annotations

import copy
import logging
from typing import Any

import yaml
from kedro.config import MissingConfigException
from kedro.framework.project import pipelines
from kedro.framework.session import KedroSession
from kedro.framework.startup import ProjectMetadata
from kedro.pipeline import Pipeline, node

from kedro_databricks.utils import (
    TASK_KEY_ORDER,
    WORKFLOW_KEY_ORDER,
    _remove_nulls_from_dict,
    _sort_dict,
    get_entry_point,
    make_workflow_name,
    require_databricks_run_script,
    update_list,
)

DEFAULT = "default"


class BundleController:
    def __init__(
        self, metadata: ProjectMetadata, env: str, config_dir: str = None
    ) -> None:
        """Create a new instance of the BundleController.

        Args:
            metadata (ProjectMetadata): The metadata of the project
            env (str): The name of the kedro environment
            config_dir (str, optional): The name of the configuration directory. Defaults to None.
        """

        self.metadata = metadata
        self.env = env
        self._conf = config_dir
        self.project_name = metadata.project_name
        self.project_path = metadata.project_path
        self.package_name = metadata.package_name
        self.pipelines = pipelines
        self.log = logging.getLogger(self.package_name)
        self.remote_conf_dir = f"/dbfs/FileStore/{self.package_name}/{config_dir}"
        self.local_conf_dir = self.metadata.project_path / config_dir / env
        self.conf = self._load_env_config(MSG="Loading configuration")

    def _workflows_to_resources(
        self, workflows: dict[str, dict[str, Any]], MSG: str = ""
    ) -> dict[str, dict[str, Any]]:
        """Convert Databricks workflows to Databricks resources.

        Args:
            workflows (Dict[str, Dict[str, Any]]): A dictionary of Databricks workflows

        Returns:
            Dict[str, Dict[str, Any]]: A dictionary of Databricks resources
        """
        resources = {
            name: {"resources": {"jobs": {name: wf}}} for name, wf in workflows.items()
        }
        self.log.debug(resources)
        self.log.info(f"{MSG}: Databricks resources successfully generated.")
        return resources

    def generate_resources(
        self, pipeline_name: str | None = None, MSG: str = ""
    ) -> dict[str, dict[str, Any]]:
        """Generate Databricks resources for the given pipelines.

        Finds all pipelines in the project and generates Databricks asset bundle resources
        for each according to the Databricks REST API

        Args:
            env (str): The name of the kedro environment to be used by the workflow
            pipeline_name (str | None): The name of the pipeline for which Databricks asset bundle resources should be generated.
            If None, generates all pipelines.
            MSG (str): The message to display

        Returns:
            dict[str, dict[str, Any]]: A dictionary of pipeline names and their Databricks resources
        """
        workflows = {}
        pipeline = self.pipelines.get(pipeline_name)
        if pipeline:
            self.log.info(f"Generating resources for pipeline '{pipeline_name}'")
            name = make_workflow_name(self.package_name, pipeline_name)
            workflows[name] = self._create_workflow(name=name, pipeline=pipeline)
            return self._workflows_to_resources(workflows, MSG)

        for pipe_name, pipeline in self.pipelines.items():
            if len(pipeline.nodes) == 0:
                continue
            name = make_workflow_name(self.package_name, pipe_name)
            workflow = self._create_workflow(name=name, pipeline=pipeline)
            self.log.debug(f"Workflow '{name}' successfully created.")
            self.log.debug(workflow)
            workflows[name] = workflow

        return self._workflows_to_resources(workflows, MSG)

    def apply_overrides(self, resources: dict[str, Any], default_key: str = DEFAULT):
        """Apply overrides to the Databricks resources.

        Args:
            resources (Dict[str, Any]): dictionary of Databricks resources
            overrides (Dict[str, Any]): dictionary of overrides
            default_key (str, optional): default key to use for overrides

        Returns:
            Dict[str, Any]: dictionary of Databricks resources with overrides applied
        """
        default_workflow = self.conf.pop(default_key, {})
        default_tasks = default_workflow.get("tasks", [])
        default_task = _get_value_by_key(default_tasks, "task_key", default_key)

        for name, resource in resources.items():
            workflow = resource["resources"]["jobs"][name]
            workflow_overrides = copy.deepcopy(default_workflow)
            workflow_overrides.update(self.conf.get(name, {}))
            workflow_task_overrides = workflow_overrides.pop("tasks", [])
            task_overrides = _get_value_by_key(
                workflow_task_overrides, "task_key", default_key
            )
            if not task_overrides:
                task_overrides = default_task

            resources[name]["resources"]["jobs"][name] = _apply_overrides(
                workflow, workflow_overrides, default_task=task_overrides
            )

        return resources

    def _load_env_config(self, MSG: str = "") -> dict[str, Any]:
        """Load the Databricks configuration for the given environment.

        Args:
            env (str): The name of the kedro environment
            MSG (str): The message to display

        Returns:
            dict[str, Any]: The Databricks configuration for the given environment
        """
        # If the configuration directory does not exist, Kedro will not load any configuration
        if not self.local_conf_dir.exists():
            self.log.warning(
                f"{MSG}: Creating {self.local_conf_dir.relative_to(self.project_path)}"
            )
            self.local_conf_dir.mkdir(parents=True)

        with KedroSession.create(
            project_path=self.project_path, env=self.env
        ) as session:
            config_loader = session._get_config_loader()
            # Backwards compatibility for ConfigLoader that does not support `config_patterns`
            if not hasattr(config_loader, "config_patterns"):
                return config_loader.get(
                    "databricks*", "databricks/**"
                )  # pragma: no cover

            # Set the default pattern for `databricks` if not provided in `settings.py`
            if "databricks" not in config_loader.config_patterns.keys():
                config_loader.config_patterns.update(  # pragma: no cover
                    {"databricks": ["databricks*", "databricks/**"]}
                )

            assert "databricks" in config_loader.config_patterns.keys()

            # Load the config
            try:
                return config_loader["databricks"]
            except MissingConfigException:  # pragma: no cover
                self.log.warning("No Databricks configuration found.")
                return {}

    def _create_workflow(self, name: str, pipeline: Pipeline) -> dict[str, Any]:
        """Create a Databricks workflow for a given pipeline.

        Args:
            name (str): name of the pipeline
            pipeline (Pipeline): Kedro pipeline object
            env (str): name of the env to be used by the tasks of the workflow

        Returns:
            Dict[str, Any]: a Databricks workflow
        """
        ## Follows the Databricks REST API schema
        ## https://docs.databricks.com/api/workspace/jobs/create
        workflow = {
            "name": name,
            "tasks": [
                self._create_task(node.name, depends_on=deps)
                for node, deps in sorted(pipeline.node_dependencies.items())
            ],
            "format": "MULTI_TASK",
        }

        return _remove_nulls_from_dict(_sort_dict(workflow, WORKFLOW_KEY_ORDER))

    def _create_task(
        self,
        name: str,
        depends_on: list[node],
    ) -> dict[str, Any]:
        """Create a Databricks task for a given node.

        Args:
            name (str): name of the node
            depends_on (List[Node]): list of nodes that the task depends on
            env (str): name of the env to be used by the task

        Returns:
            Dict[str, Any]: a Databricks task
        """
        ## Follows the Databricks REST API schema. See "tasks" in the link below
        ## https://docs.databricks.com/api/workspace/jobs/create
        entry_point = get_entry_point(self.project_name)
        params = [
            "--nodes",
            name,
            "--conf-source",
            self.remote_conf_dir,
            "--env",
            self.env,
        ]

        if require_databricks_run_script():  # pragma: no cover
            entry_point = "databricks_run"
            params = params + ["--package-name", self.package_name]

        depends_on = sorted(list(depends_on), key=lambda dep: dep.name)
        task = {
            "task_key": name.replace(".", "_"),
            "libraries": [{"whl": "../dist/*.whl"}],
            "depends_on": [
                {"task_key": dep.name.replace(".", "_")} for dep in depends_on
            ],
            "python_wheel_task": {
                "package_name": self.package_name,
                "entry_point": entry_point,
                "parameters": params,
            },
        }

        return _sort_dict(task, TASK_KEY_ORDER)

    def save_bundled_resources(
        self, resources: dict[str, dict[str, Any]], overwrite: bool = False
    ):
        """Save the generated resources to the project directory.

        Args:
            resources (Dict[str, Dict[str, Any]]): A dictionary of pipeline names and their Databricks resources
            metadata (ProjectMetadata): The metadata of the project
            overwrite (bool): Whether to overwrite existing resources
        """
        resources_dir = self.project_path / "resources"
        resources_dir.mkdir(exist_ok=True)
        for name, resource in resources.items():
            MSG = f"Writing resource '{name}'"
            p = resources_dir / f"{name}.yml"

            if p.exists() and not overwrite:  # pragma: no cover
                self.log.warning(
                    f"{MSG}: {p.relative_to(self.project_path)} already exists."
                    " Use --overwrite to replace."
                )
                continue

            with open(p, "w") as f:
                self.log.info(f"{MSG}: Wrote {p.relative_to(self.project_path)}")
                yaml.dump(
                    resource, f, default_flow_style=False, indent=4, sort_keys=False
                )


def _apply_overrides(
    workflow: dict[str, Any],
    overrides: dict[str, Any],
    default_task: dict[str, Any] = {},
):
    from mergedeep import merge

    workflow["description"] = workflow.get("description", overrides.get("description"))
    workflow["edit_mode"] = workflow.get("edit_mode", overrides.get("edit_mode"))
    workflow["max_concurrent_runs"] = workflow.get(
        "max_concurrent_runs", overrides.get("max_concurrent_runs")
    )
    workflow["timeout_seconds"] = workflow.get(
        "timeout_seconds", overrides.get("timeout_seconds")
    )

    workflow["health"] = merge(workflow.get("health", {}), overrides.get("health", {}))
    workflow["email_notifications"] = merge(
        workflow.get("email_notifications", {}),
        overrides.get("email_notifications", {}),
    )
    workflow["webhook_notifications"] = merge(
        workflow.get("webhook_notifications", {}),
        overrides.get("webhook_notifications", {}),
    )
    workflow["notification_settings"] = merge(
        workflow.get("notification_settings", {}),
        overrides.get("notification_settings", {}),
    )
    workflow["schedule"] = merge(
        workflow.get("schedule", {}), overrides.get("schedule", {})
    )
    workflow["trigger"] = merge(
        workflow.get("trigger", {}), overrides.get("trigger", {})
    )
    workflow["continuous"] = merge(
        workflow.get("continuous", {}), overrides.get("continuous", {})
    )
    workflow["git_source"] = merge(
        workflow.get("git_source", {}), overrides.get("git_source", {})
    )
    workflow["tags"] = merge(workflow.get("tags", {}), overrides.get("tags", {}))
    workflow["queue"] = merge(workflow.get("queue", {}), overrides.get("queue", {}))
    workflow["run_as"] = merge(workflow.get("run_as", {}), overrides.get("run_as", {}))
    workflow["deployment"] = merge(
        workflow.get("deployment", {}), overrides.get("deployment", {})
    )

    workflow["access_control_list"] = merge(
        workflow.get("access_control_list", {}),
        overrides.get("access_control_list", {}),
    )

    workflow["tasks"] = update_list(
        workflow.get("tasks", []), overrides.get("tasks", []), "task_key", default_task
    )
    workflow["job_clusters"] = update_list(
        workflow.get("job_clusters", []),
        overrides.get("job_clusters", []),
        "job_cluster_key",
    )
    workflow["parameters"] = update_list(
        workflow.get("parameters", []), overrides.get("parameters", []), "name"
    )
    workflow["environments"] = update_list(
        workflow.get("environments", []),
        overrides.get("environments", []),
        "environment_key",
    )

    workflow["format"] = "MULTI_TASK"

    return _remove_nulls_from_dict(_sort_dict(workflow, WORKFLOW_KEY_ORDER))


def _get_value_by_key(
    lst: list[dict[str, Any]], lookup: str, key: str
) -> dict[str, Any]:
    result = {}
    for d in lst:
        if d.get(lookup) == key:
            result = copy.deepcopy(d)
            break
    if result.get(lookup):
        result.pop(lookup)
    return result



from kedro.framework.startup import ProjectMetadata


def reset_init(metadata: ProjectMetadata):
    (metadata.project_path / "databricks.yml").unlink(missing_ok=True)
    (metadata.project_path / "conf" / "base" / "databricks.yml").unlink(missing_ok=True)



from __future__ import annotations

import logging
import shutil

import click
from kedro.framework.cli.utils import ENV_HELP
from kedro.framework.startup import ProjectMetadata

from kedro_databricks.bundle import (
    BundleController,
)
from kedro_databricks.deploy import DeployController
from kedro_databricks.init import InitController
from kedro_databricks.utils import require_databricks_run_script

DEFAULT_RUN_ENV = "local"
DEFAULT_CONF_FOLDER = "conf"
DEFAULT_CONFIG_KEY = "default"
DEFAULT_CONFIG_HELP = "Set the key for the default configuration"
CONF_HELP = "Set the conf folder. Default to `conf`."
_PROVIDER_PROMPT = """
Please select your cloud provider:
1. Azure
2. AWS
3. GCP
"""
_PROVIDER_MAP = {
    "1": "azure",
    "2": "aws",
    "3": "gcp",
}


@click.group(name="Kedro-Databricks")
def commands():
    pass


@commands.group(name="databricks")
def databricks_commands():
    """Run project with Databricks"""
    pass


@databricks_commands.command()
@click.option("-d", "--default", default=DEFAULT_CONFIG_KEY, help=DEFAULT_CONFIG_HELP)
@click.option("--provider", prompt=_PROVIDER_PROMPT, default="1")
@click.pass_obj
def init(
    metadata: ProjectMetadata,
    default: str,
    provider: str,
):
    """Initialize Databricks Asset Bundle configuration"""
    provider_name = _PROVIDER_MAP.get(provider)
    controller = InitController(metadata)
    controller.bundle_init()
    controller.write_kedro_databricks_config(default, provider_name)
    if require_databricks_run_script():  # pragma: no cover
        log = logging.getLogger(metadata.package_name)
        log.warning(
            "Kedro version less than 0.19.8 requires a script to run tasks on Databricks. "
        )
        controller.write_databricks_run_script()
    controller.substitute_catalog_paths()


@databricks_commands.command()
@click.option("-d", "--default", default=DEFAULT_CONFIG_KEY, help=DEFAULT_CONFIG_HELP)
@click.option("-e", "--env", default=DEFAULT_RUN_ENV, help=ENV_HELP)
@click.option("-c", "--conf", default=DEFAULT_CONF_FOLDER, help=CONF_HELP)
@click.option("-p", "--pipeline", default=None, help="Bundle a single pipeline")
@click.option(
    "--overwrite",
    default=False,
    is_flag=True,
    show_default=True,
    help="Overwrite the existing resources",
)
@click.pass_obj
def bundle(
    metadata: ProjectMetadata,
    default: str,
    env: str,
    conf: str,
    pipeline: str | None,
    overwrite: bool,
):
    """Convert kedro pipelines into Databricks asset bundle resources"""
    if default.startswith("_"):  # pragma: no cover
        raise ValueError(
            "Default key cannot start with `_` as this is not recognized by OmegaConf."
        )

    MSG = "Create Asset Bundle Resources"
    controller = BundleController(metadata, env, conf)
    resources = controller.generate_resources(pipeline, MSG)
    bundle_resources = controller.apply_overrides(resources, default)
    controller.save_bundled_resources(bundle_resources, overwrite)


@databricks_commands.command()
@click.option("-e", "--env", default=DEFAULT_RUN_ENV, help=ENV_HELP)
@click.option(
    "-t",
    "--target",
    default=None,
    help="Databricks target environment. Defaults to the `env` value.",
)
@click.option(
    "-b",
    "--bundle/--no-bundle",
    default=False,
    help="Bundle the project before deploying",
)
@click.option("-c", "--conf", default=DEFAULT_CONF_FOLDER, help=CONF_HELP)
@click.option("-d", "--debug/--no-debug", default=False, help="Enable debug mode")
@click.option("-p", "--pipeline", default=None, help="Bundle a single pipeline")
@click.pass_obj
def deploy(
    metadata: ProjectMetadata,
    env: str,
    target: str | None,
    bundle: bool,
    conf: str,
    pipeline: str,
    debug: bool,
):
    """Deploy the asset bundle to Databricks"""

    if shutil.which("databricks") is None:  # pragma: no cover
        raise Exception("databricks CLI is not installed")
    controller = DeployController(metadata)
    controller.go_to_project()
    controller.validate_databricks_config()
    controller.build_project()
    if bundle is True:
        bundle_controller = BundleController(metadata, env, conf)
        workflows = bundle_controller.generate_resources(pipeline)
        bundle_resources = bundle_controller.apply_overrides(workflows, "default")
        bundle_controller.save_bundled_resources(bundle_resources, overwrite=True)
    controller.create_dbfs_dir()
    controller.upload_project_config(conf)
    controller.upload_project_data()
    if target is None:
        target = env
    controller.deploy_project(target=target, debug=debug)



from __future__ import annotations

import logging
import os
import tarfile
from collections import namedtuple
from pathlib import Path

from databricks.sdk import WorkspaceClient
from kedro.framework.project import _ProjectPipelines
from kedro.framework.project import pipelines as kedro_pipelines
from kedro.framework.startup import ProjectMetadata

from kedro_databricks.utils import make_workflow_name, run_cmd

_INVALID_CONFIG_MSG = """
No `databricks.yml` file found. Maybe you forgot to initialize the Databricks bundle?

You can initialize the Databricks bundle by running:

```
kedro databricks init
```
"""

JobLink = namedtuple("JobLink", ["name", "url", "is_dev"])


class DeployController:
    def __init__(self, metadata: ProjectMetadata) -> None:
        self._msg = "Deploying to Databricks"
        self.metadata = metadata
        self.package_name = metadata.package_name
        self.project_path = metadata.project_path
        self.log = logging.getLogger(metadata.package_name)

    def go_to_project(self) -> Path:
        """Change the current working directory to the project path.

        Returns:
            pathlib.Path: Path to the project directory.

        Raises:
            FileNotFoundError: If the project path does not exist.
        """
        project_path = Path(self.project_path)
        if not project_path.exists():
            raise FileNotFoundError(f"Project path {project_path} does not exist")
        os.chdir(project_path)
        return project_path

    def validate_databricks_config(self):
        """Check if the Databricks configuration file exists.

        Returns:
            bool: Whether the Databricks configuration file exists.

        Raises:
            FileNotFoundError: If the Databricks configuration file does not exist.
        """
        if not (self.project_path / "databricks.yml").exists():
            raise FileNotFoundError(_INVALID_CONFIG_MSG)
        return True

    def create_dbfs_dir(self):  # pragma: no cover
        """Create a directory in DBFS."""
        run_cmd(
            ["databricks", "fs", "mkdirs", f"dbfs:/FileStore/{self.package_name}"],
            msg=self._msg,
            warn=True,
        )

    def upload_project_data(self):  # pragma: no cover
        """Upload the project data to DBFS.

        Args:
            metadata (ProjectMetadata): Project metadata.
            MSG (str): Message to display.
        """
        target_path = f"dbfs:/FileStore/{self.package_name}/data"
        source_path = self.project_path / "data"
        if not source_path.exists():
            self.log.warning(f"Data path {source_path} does not exist")
            return

        self.log.info(
            f"{self._msg}: Uploading {source_path.relative_to(self.project_path)} to {target_path}"
        )
        run_cmd(
            [
                "databricks",
                "fs",
                "cp",
                "-r",
                "--overwrite",
                source_path.as_posix(),
                target_path,
            ],
            msg=self._msg,
        )
        self.log.info(f"{self._msg}: Data uploaded to {target_path}")

    def upload_project_config(self, conf: str):  # pragma: no cover
        """Upload the project configuration to DBFS.

        Args:
            conf (str): The conf folder.
        """
        conf_tar = self.project_path / f"dist/conf-{self.package_name}.tar.gz"
        with tarfile.open(conf_tar) as f:
            f.extractall("dist/", filter="tar")

        target_path = f"dbfs:/FileStore/{self.package_name}/{conf}"
        source_path = self.project_path / "dist" / conf
        if not source_path.exists():
            raise FileNotFoundError(f"Configuration path {source_path} does not exist")

        self.log.info(f"{self._msg}: Uploading configuration to {target_path}")
        run_cmd(
            [
                "databricks",
                "fs",
                "cp",
                "-r",
                "--overwrite",
                source_path.as_posix(),
                target_path,
            ],
            msg=self._msg,
        )
        self.log.info(f"{self._msg}: Configuration uploaded to {target_path}")

    def build_project(self):  # pragma: no cover
        """Build the project."""
        self.log.info(f"{self._msg}: Building the project")
        self.go_to_project()
        build_cmd = ["kedro", "package"]
        result = run_cmd(build_cmd, msg=self._msg)
        return result

    def deploy_project(self, target: str, debug: bool = False):
        """Deploy the project to Databricks.

        Args:
            target (str): Databricks target environment to deploy to.
            debug (bool): Whether to enable debug mode.
        """
        self.log.info(
            f"{self._msg}: Running `databricks bundle deploy --target {target}`"
        )
        deploy_cmd = ["databricks", "bundle", "deploy", "--target", target]
        if debug:
            deploy_cmd.append("--debug")
        run_cmd(deploy_cmd, msg=self._msg)
        self.log_deployed_resources(only_dev=target in ["dev", "local"])

    def log_deployed_resources(
        self, pipelines: _ProjectPipelines = kedro_pipelines, only_dev: bool = False
    ) -> dict[str, set[str]]:
        """Print the pipelines."""
        w = WorkspaceClient()
        job_host = f"{w.config.host}/jobs"
        username = w.current_user.me().user_name.split("@")[0]
        all_jobs = {job.settings.name: job for job in w.jobs.list()}
        jobs = self._gather_user_jobs(all_jobs, pipelines, username, job_host)
        self.log.info(f"{self._msg}: Successfully Deployed Jobs")
        for job in jobs:
            if only_dev and not job.is_dev:
                continue
            self.log.info(f"Run '{job.name}' at {job.url}")
        return jobs

    def _gather_user_jobs(
        self,
        all_jobs: dict[str, str],
        pipelines: _ProjectPipelines,
        username,
        job_host,
    ) -> set[JobLink]:
        jobs = set()
        for job_name, job in all_jobs.items():
            is_dev = job_name.startswith("[dev")
            is_valid = self._is_valid_job(pipelines, job_name)
            if (is_dev and username not in job_name) or not is_valid:
                continue
            n = job_name.split(" - ")[0]
            link = JobLink(name=n, url=f"{job_host}/{job.job_id}", is_dev=is_dev)
            jobs.add(link)
        return jobs

    def _is_valid_job(self, pipelines: _ProjectPipelines, job_name: str) -> bool:
        return any(
            make_workflow_name(self.package_name, pipeline_name) in job_name
            for pipeline_name in pipelines
        )



import yaml

from kedro_databricks.plugin import commands


def test_databricks_bundle_fail(cli_runner, metadata):
    bundle_fail = ["databricks", "bundle", "--default", "_deault"]
    result = cli_runner.invoke(commands, bundle_fail, obj=metadata)
    assert result.exit_code == 1, (result.exit_code, result.stdout)


def test_databricks_bundle_with_overrides(kedro_project, cli_runner, metadata):
    init_cmd = ["databricks", "init"]
    result = cli_runner.invoke(commands, init_cmd, obj=metadata)
    override_path = metadata.project_path / "conf" / "base" / "databricks.yml"
    assert result.exit_code == 0, (result.exit_code, result.stdout)
    assert override_path.exists(), "Override file not created"

    command = ["databricks", "bundle", "--env", "dev"]
    result = cli_runner.invoke(commands, command, obj=metadata)
    resource_dir = kedro_project / "resources"
    conf_dir = kedro_project / "conf" / "dev"
    assert result.exit_code == 0, (result.exit_code, result.stdout)
    assert resource_dir.exists(), "Resource directory not created"
    assert resource_dir.is_dir(), "Resource directory is not a directory"
    assert conf_dir.exists(), "Configuration directory not created"

    files = [p.name for p in resource_dir.rglob("*")]
    files.sort()
    assert files == [
        f"{metadata.package_name}.yml",
        f"{metadata.package_name}_ds.yml",
    ], f"Resource files not created: {', '.join(files)}"

    resources_with_overrides = []
    for p in files:
        with open(resource_dir / p) as f:
            resource = yaml.safe_load(f)
            resources_with_overrides.append(resource)

    for resource, file_name in zip(resources_with_overrides, files):
        assert resource.get("resources") is not None

        jobs = resource["resources"]["jobs"]
        assert len(jobs) == 1

        job = jobs.get(file_name.split(".")[0])
        assert job is not None

        tasks = job.get("tasks")
        assert tasks is not None
        assert len(tasks) == 5

        for i, task in enumerate(tasks):
            assert task.get("task_key") == f"node{i}"
            assert task.get("job_cluster_key") == "default"
            params = task.get("python_wheel_task").get("parameters")
            for j, param in enumerate(params):
                if param == "--env":
                    assert params[j + 1] == "dev"


def test_databricks_bundle_with_conf(kedro_project, cli_runner, metadata):
    """Test the `bundle` command"""

    init_cmd = ["databricks", "init"]
    result = cli_runner.invoke(commands, init_cmd, obj=metadata)
    override_path = (
        metadata.project_path / "conf" / "sub_pipeline" / "base" / "databricks.yml"
    )
    assert result.exit_code == 0, (result.exit_code, result.stdout)
    override_path.unlink(missing_ok=True)
    assert not override_path.exists(), "Override file not created"

    command = ["databricks", "bundle", "--env", "dev", "--conf", "conf/sub_pipeline"]
    result = cli_runner.invoke(commands, command, obj=metadata)
    resource_dir = kedro_project / "resources"
    conf_dir = kedro_project / "conf" / "sub_pipeline" / "dev"
    assert result.exit_code == 0, (result.exit_code, result.stdout)
    assert resource_dir.exists(), "Resource directory not created"
    assert resource_dir.is_dir(), "Resource directory is not a directory"
    assert conf_dir.exists(), "Configuration directory not created"

    files = [p.name for p in resource_dir.rglob("*")]
    files.sort()
    assert files == [
        f"{metadata.package_name}.yml",
        f"{metadata.package_name}_ds.yml",
    ], f"Resource files not created: {', '.join(files)}"

    resources_without_overrides = []
    for p in files:
        with open(resource_dir / p) as f:
            resource = yaml.safe_load(f)
            resources_without_overrides.append(resource)

    for resource, file_name in zip(resources_without_overrides, files):
        assert resource.get("resources") is not None

        jobs = resource["resources"]["jobs"]
        assert len(jobs) == 1

        job = jobs.get(file_name.split(".")[0])
        assert job is not None

        tasks = job.get("tasks")
        assert tasks is not None
        assert len(tasks) == 5

        for i, task in enumerate(tasks):
            assert task.get("task_key") == f"node{i}"
            params = task.get("python_wheel_task").get("parameters")
            for j, param in enumerate(params):
                if param == "--env":
                    assert params[j + 1] == "dev"


def test_databricks_bundle_without_overrides(kedro_project, cli_runner, metadata):
    """Test the `bundle` command"""

    init_cmd = ["databricks", "init"]
    result = cli_runner.invoke(commands, init_cmd, obj=metadata)
    override_path = metadata.project_path / "conf" / "base" / "databricks.yml"
    assert result.exit_code == 0, (result.exit_code, result.stdout)
    override_path.unlink(missing_ok=True)
    assert not override_path.exists(), "Override file not created"

    command = ["databricks", "bundle", "--env", "dev"]
    result = cli_runner.invoke(commands, command, obj=metadata)
    resource_dir = kedro_project / "resources"
    conf_dir = kedro_project / "conf" / "dev"
    assert result.exit_code == 0, (result.exit_code, result.stdout)
    assert resource_dir.exists(), "Resource directory not created"
    assert resource_dir.is_dir(), "Resource directory is not a directory"
    assert conf_dir.exists(), "Configuration directory not created"

    files = [p.name for p in resource_dir.rglob("*")]
    files.sort()
    assert files == [
        f"{metadata.package_name}.yml",
        f"{metadata.package_name}_ds.yml",
    ], f"Resource files not created: {', '.join(files)}"

    resources_without_overrides = []
    for p in files:
        with open(resource_dir / p) as f:
            resource = yaml.safe_load(f)
            resources_without_overrides.append(resource)

    for resource, file_name in zip(resources_without_overrides, files):
        assert resource.get("resources") is not None

        jobs = resource["resources"]["jobs"]
        assert len(jobs) == 1

        job = jobs.get(file_name.split(".")[0])
        assert job is not None

        tasks = job.get("tasks")
        assert tasks is not None
        assert len(tasks) == 5

        for i, task in enumerate(tasks):
            assert task.get("task_key") == f"node{i}"
            params = task.get("python_wheel_task").get("parameters")
            for j, param in enumerate(params):
                if param == "--env":
                    assert params[j + 1] == "dev"



"""
This file contains the fixtures that are reusable by any tests within
this directory. You don't need to import the fixtures as pytest will
discover them automatically. More info here:
https://docs.pytest.org/en/latest/fixture.html
"""

from __future__ import annotations

import os
from pathlib import Path

from click.testing import CliRunner
from kedro.framework.cli.starters import create_cli as kedro_cli
from kedro.framework.startup import bootstrap_project
from pytest import fixture


@fixture(name="cli_runner", scope="session")
def cli_runner():
    runner = CliRunner()
    with runner.isolated_filesystem():
        yield runner


def _create_kedro_settings_py(file_name: Path, patterns: list[str]):
    patterns_str = ", ".join([f'"{p}"' for p in patterns])
    content = f"""CONFIG_LOADER_ARGS = {{
    "base_env": "base",
    "default_run_env": "local",
    "config_patterns": {{
        "databricks": [{patterns_str}],  # configure the pattern for configuration files
    }}
}}
"""
    file_name.write_text(content)


@fixture(scope="session")
def kedro_project(cli_runner):
    cli_runner.invoke(
        # Supply name, tools, and example to skip interactive prompts
        kedro_cli,
        [
            "new",
            "-v",
            "--name",
            "Fake Project",
            "--tools",
            "none",
            "--example",
            "no",
        ],
    )
    pipeline_registry_py = """
from kedro.pipeline import Pipeline, node


def identity(arg):
    return arg


def register_pipelines():
    pipeline = Pipeline(
        [
            node(identity, ["input"], ["intermediate"], name="node0", tags=["tag0", "tag1"]),
            node(identity, ["intermediate"], ["output"], name="node1"),
            node(identity, ["intermediate"], ["output2"], name="node2", tags=["tag0"]),
            node(identity, ["intermediate"], ["output3"], name="node3", tags=["tag1", "tag2"]),
            node(identity, ["intermediate"], ["output4"], name="node4", tags=["tag2"]),
        ],
        tags="pipeline0",
    )
    return {
        "__default__": pipeline,
        "ds": pipeline,
    }
    """

    project_path = Path().cwd() / "fake-project"
    (project_path / "src" / "fake_project" / "pipeline_registry.py").write_text(
        pipeline_registry_py
    )

    settings_file = project_path / "src" / "fake_project" / "settings.py"
    _create_kedro_settings_py(settings_file, ["databricks*", "databricks/**"])

    os.chdir(project_path)
    return project_path


@fixture(scope="session")
def metadata(kedro_project):
    # cwd() depends on ^ the isolated filesystem, created by CliRunner()
    project_path = kedro_project.resolve()
    metadata = bootstrap_project(project_path)
    return metadata



from __future__ import annotations

import copy
import logging
import re
import shutil
import subprocess
from typing import Any

from kedro import __version__ as kedro_version

KEDRO_VERSION = [int(x) for x in kedro_version.split(".")]
TASK_KEY_ORDER = [
    "task_key",
    "job_cluster_key",
    "new_cluster",
    "depends_on",
    "spark_python_task",
    "python_wheel_task",
]

WORKFLOW_KEY_ORDER = [
    "name",
    "tags",
    "access_control_list",
    "email_notifications",
    "schedule",
    "max_concurrent_runs",
    "job_clusters",
    "tasks",
]


def has_databricks_cli() -> bool:
    """Check if the Databricks CLI is installed."""
    return shutil.which("databricks") is not None


def get_entry_point(project_name: str) -> str:
    """Get the entry point for a project.

    Args:
        project_name (str): name of the project

    Returns:
        str: entry point for the project
    """
    entrypoint = project_name.strip().lower()
    entrypoint = re.sub(r" +", " ", entrypoint)
    return re.sub(r"[^a-zA-Z]", "-", entrypoint)


def require_databricks_run_script() -> bool:
    """Check if the current Kedro version is less than 0.19.8.

    Kedro 0.19.8 introduced a new `run_script` method that is required for
    running tasks on Databricks. This method is not available in earlier
    versions of Kedro. This function checks if the current Kedro version is
    less than 0.19.8.

    Returns:
        bool: whether the current Kedro version is less than 0.19.8
    """
    return KEDRO_VERSION < [0, 19, 8]


def run_cmd(
    cmd: list[str], msg: str = "Failed to run command", warn: bool = False
) -> subprocess.CompletedProcess | None:
    """Run a shell command.

    Args:
        cmds (List[str]): list of commands to run
        msg (str, optional): message to raise if the command fails
        warn (bool): whether to log a warning if the command fails
    """

    try:
        result = subprocess.run(cmd, stdout=subprocess.PIPE, check=True)
        for line in result.stdout.decode().split("\n"):
            logging.info(line)
        return result
    except Exception as e:
        if warn:
            logging.warning(f"{msg}: {e}")
            return None
        else:
            raise Exception(f"{msg}: {e}")


def make_workflow_name(package_name, pipeline_name: str) -> str:
    """Create a name for the Databricks workflow.

    Args:
        pipeline_name (str): The name of the pipeline

    Returns:
        str: The name of the workflow
    """
    if pipeline_name == "__default__":
        return package_name
    return f"{package_name}_{pipeline_name}"


def update_list(
    old: list[dict[str, Any]],
    new: list[dict[str, Any]],
    lookup_key: str,
    default: dict[str, Any] = {},
):
    """Update a list of dictionaries with another list of dictionaries.

    Args:
        old (List[Dict[str, Any]]): list of dictionaries to update
        new (List[Dict[str, Any]]): list of dictionaries to update with
        lookup_key (str): key to use for looking up dictionaries
        default (Dict[str, Any], optional): default dictionary to use for updating

    Returns:
        List[Dict[str, Any]]: updated list of dictionaries
    """
    assert isinstance(
        old, list
    ), f"old must be a list not {type(old)} for key: {lookup_key} - {old}"
    assert isinstance(
        new, list
    ), f"new must be a list not {type(new)} for key: {lookup_key} - {new}"
    from mergedeep import merge

    old_obj = {curr.pop(lookup_key): curr for curr in old}
    new_obj = {update.pop(lookup_key): update for update in new}
    keys = set(old_obj.keys()).union(set(new_obj.keys()))

    for key in keys:
        update = copy.deepcopy(default)
        update.update(new_obj.get(key, {}))
        new = merge(old_obj.get(key, {}), update)
        old_obj[key] = new

    return [{lookup_key: k, **v} for k, v in old_obj.items()]


def _sort_dict(d: dict[Any, Any], key_order: list[str]) -> dict[Any, Any]:
    """Recursively sort the keys of a dictionary.

    Args:
        d (Dict[Any, Any]): dictionary to sort
        key_order (List[str]): list of keys to sort by

    Returns:
        Dict[Any, Any]: dictionary with ordered values
    """
    other_keys = [k for k in d.keys() if k not in key_order]
    order = key_order + other_keys

    return dict(sorted(d.items(), key=lambda x: order.index(x[0])))


def _is_null_or_empty(x: Any) -> bool:
    """Check if a value is None or an empty dictionary.

    Args:
        x (Any): value to check

    Returns:
        bool: whether the value is None or an empty dictionary
    """
    return x is None or (isinstance(x, (dict, list)) and len(x) == 0)


def _remove_nulls_from_list(
    lst: list[dict | float | int | str | bool],
) -> list[dict | list]:
    """Remove None values from a list.

    Args:
        l (List[Dict[Any, Any]]): list to remove None values from

    Returns:
        List[Dict[Any, Any]]: list with None values removed
    """
    for i, item in enumerate(lst):
        value = remove_nulls(item)
        if _is_null_or_empty(value):
            lst.remove(item)
        else:
            lst[i] = value


def _remove_nulls_from_dict(
    d: dict[str, Any],
) -> dict[str, float | int | str | bool]:
    """Remove None values from a dictionary.

    Args:
        d (Dict[str, Any]): dictionary to remove None values from

    Returns:
        Dict[str, float | int | str | bool]: dictionary with None values removed
    """
    for k, v in list(d.items()):
        value = remove_nulls(v)
        if _is_null_or_empty(value):
            del d[k]
        else:
            d[k] = value
    return d


def remove_nulls(value: dict | list) -> dict | list:
    """Remove None values from a dictionary or list.

    Args:
        value (Dict[Any, Any] | List[Dict[Any, Any]]): dictionary or list to remove None values from

    Returns:
        Dict[Any, Any] | List[Dict[Any, Any]]: dictionary or list with None values removed
    """
    non_null = copy.deepcopy(value)
    if isinstance(non_null, dict):
        _remove_nulls_from_dict(non_null)
    elif isinstance(non_null, list):
        _remove_nulls_from_list(non_null)
    return non_null



from __future__ import annotations

import copy

from kedro.pipeline import Pipeline, node

from kedro_databricks.bundle import BundleController
from kedro_databricks.utils import require_databricks_run_script


def identity(arg):
    return arg


pipeline = Pipeline(
    [
        node(
            identity,
            ["input"],
            ["intermediate"],
            name="node0",
            tags=["tag0", "tag1"],
        ),
        node(identity, ["intermediate"], ["output"], name="node1"),
        node(identity, ["intermediate"], ["output2"], name="node2", tags=["tag0"]),
        node(
            identity,
            ["intermediate"],
            ["output3"],
            name="node3",
            tags=["tag1", "tag2"],
        ),
        node(identity, ["intermediate"], ["output4"], name="node4", tags=["tag2"]),
    ]
)


def _generate_task(task_key: int | str, depends_on: list[str] = [], conf: str = "conf"):
    entry_point = "fake-project"
    params = [
        "--nodes",
        task_key,
        "--conf-source",
        f"/dbfs/FileStore/fake_project/{conf}",
        "--env",
        "fake_env",
    ]

    if require_databricks_run_script():
        entry_point = "databricks_run"
        params = params + ["--package-name", "fake_project"]

    task = {
        "task_key": task_key,
        "libraries": [
            {"whl": "../dist/*.whl"},
        ],
        "python_wheel_task": {
            "package_name": "fake_project",
            "entry_point": entry_point,
            "parameters": params,
        },
    }

    if len(depends_on) > 0:
        task["depends_on"] = [{"task_key": dep} for dep in depends_on]

    return task


def generate_workflow(conf="conf"):
    tasks = []

    for i in range(5):
        if i == 0:
            depends_on = []
        else:
            depends_on = ["node0"]
        tasks.append(_generate_task(f"node{i}", depends_on, conf))

    return {
        "name": "workflow1",
        "tasks": tasks,
        "format": "MULTI_TASK",
    }


WORKFLOW = generate_workflow()

OVERRIDES = {
    "default": {
        "job_clusters": [
            {
                "job_cluster_key": "default",
                "new_cluster": {
                    "spark_version": "7.3.x-scala2.12",
                    "node_type_id": "Standard_DS3_v2",
                    "num_workers": 2,
                    "spark_env_vars": {
                        "KEDRO_LOGGING_CONFIG": "dbfs:/path/to/logging.yml"
                    },
                },
            }
        ],
        "tasks": [
            {
                "task_key": "default",
                "job_cluster_key": "default",
            }
        ],
    }
}

WORKFLOW_OVERRIDES = {
    WORKFLOW["name"]: {
        "job_clusters": [
            {
                "job_cluster_key": "default",
                "new_cluster": {
                    "spark_version": "7.3.x-scala2.12",
                    "node_type_id": "Standard_DS3_v2",
                    "num_workers": 2,
                    "spark_env_vars": {
                        "KEDRO_LOGGING_CONFIG": "dbfs:/path/to/logging.yml"
                    },
                },
            }
        ],
        "tasks": [
            {
                "task_key": "default",
                "job_cluster_key": "default",
            }
        ],
    }
}

MIX_OVERRIDES = {
    "default": {
        "job_clusters": [
            {
                "job_cluster_key": "default",
                "new_cluster": {
                    "spark_version": "7.3.x-scala2.12",
                    "node_type_id": "Standard_DS3_v2",
                    "num_workers": 2,
                    "spark_env_vars": {
                        "KEDRO_LOGGING_CONFIG": "dbfs:/path/to/logging.yml"
                    },
                },
            }
        ]
    },
    WORKFLOW["name"]: {
        "tasks": [
            {
                "task_key": "default",
                "job_cluster_key": "default",
            }
        ],
    },
}


def _generate_testdata():
    result = copy.deepcopy(WORKFLOW)
    result["job_clusters"] = [
        {
            "job_cluster_key": "default",
            "new_cluster": {
                "spark_version": "7.3.x-scala2.12",
                "node_type_id": "Standard_DS3_v2",
                "num_workers": 2,
                "spark_env_vars": {"KEDRO_LOGGING_CONFIG": "dbfs:/path/to/logging.yml"},
            },
        }
    ]
    for task in result["tasks"]:
        task["job_cluster_key"] = "default"

    resources = {
        WORKFLOW["name"]: {
            "resources": {"jobs": {WORKFLOW["name"]: copy.deepcopy(WORKFLOW)}}
        }
    }

    return resources, result


def test_apply_resource_overrides(metadata):
    resources, result = _generate_testdata()
    controller = BundleController(metadata, "fake_env", "conf")
    controller.conf = OVERRIDES
    assert controller.apply_overrides(resources, "default") == {
        "workflow1": {"resources": {"jobs": {"workflow1": result}}}
    }, "Failed to apply default overrides"
    controller.conf = WORKFLOW_OVERRIDES
    assert controller.apply_overrides(resources, "default") == {
        "workflow1": {"resources": {"jobs": {"workflow1": result}}}
    }, "Failed to apply workflow overrides"
    controller.conf = MIX_OVERRIDES
    assert controller.apply_overrides(resources, "default") == {
        "workflow1": {"resources": {"jobs": {"workflow1": result}}}
    }, "Failed to apply mixed overrides"


def test_generate_workflow(metadata):
    controller = BundleController(metadata, "fake_env", "conf")
    assert controller._create_workflow("workflow1", pipeline) == WORKFLOW


def test_create_task(metadata):
    controller = BundleController(metadata, "fake_env", "conf")
    expected_task = _generate_task("task", ["a", "b"])
    assert controller._create_task("task", [
        node(identity, ["input"], ["output"], name="b"),
        node(identity, ["input"], ["output"], name="a")]) == expected_task


def test_generate_resources(metadata):
    controller = BundleController(metadata, "fake_env", "conf")
    controller.pipelines = {"__default__": Pipeline([])}
    assert controller.generate_resources(pipeline_name=None, MSG="Test MSG") == {}
    controller.pipelines = {
        "__default__": Pipeline([node(identity, ["input"], ["output"], name="node")])
    }
    assert controller.generate_resources(pipeline_name=None, MSG="Test MSG") == {
        "fake_project": {
            "resources": {
                "jobs": {
                    "fake_project": {
                        "format": "MULTI_TASK",
                        "name": "fake_project",
                        "tasks": [
                            _generate_task("node"),
                        ],
                    },
                },
            },
        },
    }


def test_generate_resources_another_conf(metadata):
    controller = BundleController(metadata, "fake_env", "sub_conf")
    controller.pipelines = {
        "__default__": Pipeline([node(identity, ["input"], ["output"], name="node")])
    }

    assert controller.generate_resources(pipeline_name=None, MSG="Test MSG") == {
        "fake_project": {
            "resources": {
                "jobs": {
                    "fake_project": {
                        "format": "MULTI_TASK",
                        "name": "fake_project",
                        "tasks": [
                            _generate_task("node", conf="sub_conf"),
                        ],
                    },
                },
            },
        },
    }


def test_generate_resources_in_a_sorted_manner(metadata):
    controller = BundleController(metadata, "fake_env", "conf")
    controller.pipelines = {
        "__default__": Pipeline(
            [
                node(identity, ["input"], ["b_output"], name="b_node"),
                node(identity, ["input"], ["a_output"], name="a_node"),
            ]
        )
    }
    assert controller.generate_resources(pipeline_name=None, MSG="Test MSG") == {
        "fake_project": {
            "resources": {
                "jobs": {
                    "fake_project": {
                        "format": "MULTI_TASK",
                        "name": "fake_project",
                        "tasks": [
                            _generate_task("a_node"),
                            _generate_task("b_node"),
                        ],
                    },
                },
            },
        },
    }


def test_generate_resources_for_a_single_pipeline(metadata):
    controller = BundleController(metadata, "fake_env", "conf")
    controller.pipelines = {
        "__default__": Pipeline(
            [
                node(identity, ["input"], ["a_output"], name="a_node"),
            ]
        ),
        "a_pipeline": Pipeline(
            [
                node(identity, ["input"], ["a_output"], name="a_node"),
            ]
        ),
        "b_pipeline": Pipeline(
            [
                node(identity, ["input"], ["b_output"], name="b_node"),
            ]
        ),
    }
    assert controller.generate_resources(
        pipeline_name="b_pipeline", MSG="Test MSG"
    ) == {
        "fake_project_b_pipeline": {
            "resources": {
                "jobs": {
                    "fake_project_b_pipeline": {
                        "format": "MULTI_TASK",
                        "name": "fake_project_b_pipeline",
                        "tasks": [
                            _generate_task("b_node"),
                        ],
                    },
                },
            },
        },
    }


def test_save_resoureces(metadata):
    controller = BundleController(metadata, "fake_env", "conf")

    resources = controller.generate_resources(None, "")
    bundle_resources = controller.apply_overrides(resources, "default")
    controller.save_bundled_resources(bundle_resources, True)
    resource_dir = metadata.project_path.joinpath("resources")
    assert resource_dir.exists(), "Failed to create resources directory"
    assert resource_dir.is_dir(), "resouces is not a directory"

    project_resources = resource_dir.joinpath(f"{metadata.package_name}.yml")
    project_files = ",".join([str(p) for p in resource_dir.iterdir()])
    assert (
        project_resources.exists()
    ), f"Failed to save project resources, {project_files}"



from kedro_databricks.plugin import commands
from tests.utils import reset_init


def test_databricks_init(kedro_project, cli_runner, metadata):
    """Test the `init` command"""
    reset_init(metadata)
    command = ["databricks", "init"]
    result = cli_runner.invoke(commands, command, obj=metadata)

    files = [f"{f.parent.name}/{f.name}" for f in kedro_project.rglob("*")]
    assert len(files) > 0, "Found no files in the directory."

    config_path = kedro_project / "databricks.yml"
    override_path = kedro_project / "conf" / "base" / "databricks.yml"
    assert result.exit_code == 0, (result.exit_code, result.stdout)
    assert config_path.exists(), f"Configuration at {config_path} does not exist"
    assert (
        override_path.exists()
    ), f"Resource Overrides at {override_path} does not exist"

    command = ["databricks", "init"]
    result = cli_runner.invoke(commands, command, obj=metadata)
    assert result.exit_code == 0, (result.exit_code, result.stdout)



from __future__ import annotations

import subprocess

import pytest

from kedro_databricks.utils import _is_null_or_empty, remove_nulls, run_cmd


def test_run_cmd():
    with pytest.raises(Exception) as e:
        cmd = ["ls", "non_existent_file"]
        run_cmd(cmd)
        assert "Failed to run command" in str(
            e.value
        ), f"Failed to raise exception: {cmd}"
        raise e

    with pytest.raises(Exception) as e:
        cmd = ["ls", "non_existent_file"]
        run_cmd(cmd, msg="Custom message")
        assert "Custom message" in str(e.value), f"Failed to raise exception: {cmd}"
        raise e

    run_cmd(["ls", "non_existent_file"], warn=True)
    run_cmd(["ls", "non_existent_file"], msg="Custom message", warn=True)

    result = run_cmd(["ls", "."])
    assert isinstance(result, subprocess.CompletedProcess)

    result = run_cmd(["ls", "non_existent_file"], msg="Custom message", warn=True)
    assert result is None, result


def test_remove_nulls_from_dict():
    a = {
        "a": 1,
        "b": None,
        "c": {"d": None},
        "e": {"f": {"g": None}},
        "h": {"i": {"j": {}}},
        "k": [],
        "l": [1, 2, 3],
        "m": [1, None, 3],
        "n": [1, {"o": None}, 3],
        "o": [1, [None], 3],
        "p": [1, {"q": {"r": None}}, 3],
    }

    assert remove_nulls(a) == {
        "a": 1,
        "l": [1, 2, 3],
        "m": [1, 3],
        "n": [1, 3],
        "o": [1, 3],
        "p": [1, 3],
    }, remove_nulls(a)


def test_is_null_or_empty():
    assert _is_null_or_empty(None), "Failed to check None"
    assert _is_null_or_empty({}), "Failed to check empty dict"
    assert _is_null_or_empty([]), "Failed to check empty list"
    assert not _is_null_or_empty(1), "Failed to check int"
    assert not _is_null_or_empty("a"), "Failed to check str"
    assert not _is_null_or_empty({1: 1}), "Failed to check dict"
    assert not _is_null_or_empty([1]), "Failed to check list"



from __future__ import annotations

import os
from pathlib import Path

import pytest

from kedro_databricks.deploy import DeployController


class MetadataMock:
    def __init__(self, path: str, name: str):
        self.project_path = Path(path)
        self.project_name = name
        self.package_name = name
        self.source_dir = "src"
        self.env = "local"
        self.config_file = "conf/base"
        self.project_version = "0.16.0"
        self.project_description = "Test Project Description"
        self.project_author = "Test Author"
        self.project_author_email = "author@email.com"


def test_deploy_go_to_project(metadata):
    controller = DeployController(metadata)
    project_path = controller.go_to_project()
    assert os.getcwd() == str(project_path), "Failed to change to project directory"
    with pytest.raises(FileNotFoundError):
        controller = DeployController(
            MetadataMock(
                "/tmp/non_existent_path" + str(os.getpid()), "non_existent_project"
            )
        )
        controller.go_to_project()


def test_deploy_validate_databricks_config(metadata):
    controller = DeployController(metadata)
    project_path = controller.go_to_project()
    (project_path / "databricks.yml").unlink(missing_ok=True)
    with pytest.raises(FileNotFoundError):
        controller.validate_databricks_config()
    with open(project_path / "databricks.yml", "w") as f:
        f.write("")
    controller.validate_databricks_config()


def test_deploy_build_project(metadata):
    controller = DeployController(metadata)
    result = controller.build_project()
    assert result.returncode == 0, (result.returncode, result.stdout)



from __future__ import annotations

import json
import logging
import re
import shutil
import subprocess
import tempfile
from importlib import resources
from pathlib import Path

import tomlkit
from kedro.framework.startup import ProjectMetadata

from kedro_databricks.utils import has_databricks_cli, run_cmd

NODE_TYPE_MAP = {
    "aws": "m5.xlarge",
    "azure": "Standard_DS3_v2",
    "gcp": "n1-standard-4",
}

DEFAULT_PROVIDER = "azure"
DEFAULT_NODE_TYPE_ID = NODE_TYPE_MAP.get(DEFAULT_PROVIDER, None)
TEMPLATES = resources.files("kedro_databricks").joinpath("templates")

assert NODE_TYPE_MAP is not None, "Invalid default provider"


class InitController:
    def __init__(self, metadata: ProjectMetadata) -> None:
        self.metadata = metadata
        self.project_path = metadata.project_path
        self.package_name = metadata.package_name
        self.log = logging.getLogger(self.package_name)

    def bundle_init(self):
        MSG = "Creating databricks configuration"
        if not has_databricks_cli():  # pragma: no cover
            raise Exception("databricks CLI is not installed")

        config_path = self.project_path / "databricks.yml"
        if config_path.exists():
            self.log.warning(
                f"{MSG}: {config_path.relative_to(self.project_path)} already exists."
            )
            return

        assets_dir = Path(tempfile.mkdtemp())
        shutil.copy(
            TEMPLATES / "databricks_template_schema.json",
            assets_dir / "databricks_template_schema.json",
        )

        template_dir = assets_dir / "template"
        template_dir.mkdir(exist_ok=True)
        shutil.copy(
            TEMPLATES / "databricks.yml.tmpl", template_dir / "databricks.yml.tmpl"
        )

        config = {"project_name": self.package_name, "project_slug": self.package_name}
        template_params = tempfile.NamedTemporaryFile(delete=False)
        template_params.write(json.dumps(config).encode())
        template_params.close()

        # We utilize the databricks CLI to create the bundle configuration.
        # This is a bit hacky, but it allows the plugin to tap into the authentication
        # mechanism of the databricks CLI and thereby avoid the need to store credentials
        # in the plugin.
        init_cmd = [
            "databricks",
            "bundle",
            "init",
            assets_dir.as_posix(),
            "--config-file",
            template_params.name,
            "--output-dir",
            self.project_path.as_posix(),
        ]
        try:
            run_cmd(init_cmd, msg=MSG, warn=True)
        except subprocess.CalledProcessError as e:  # pragma: no cover
            if "Asset Bundle successfully created for project" not in e.stderr.decode(
                "utf-8"
            ):
                raise e
        self.log.info(f"{MSG}: Wrote {config_path.relative_to(self.project_path)}")

        shutil.rmtree(assets_dir)

    def write_kedro_databricks_config(self, default_key: str, provider_name: str):
        MSG = "Creating bundle override configuration"
        override_path = Path(self.project_path) / "conf" / "base" / "databricks.yml"
        node_type_id = NODE_TYPE_MAP.get(provider_name, DEFAULT_NODE_TYPE_ID)
        if override_path.exists():
            self.log.warning(
                f"{MSG}: {override_path.relative_to(self.project_path)} already exists."
            )
            return

        with open(override_path, "w") as f:
            template = TEMPLATES / "kedro_databricks_config.yml.tmpl"
            f.write(
                template.read_text().format(
                    default_key=default_key,
                    package_name=self.package_name,
                    node_type_id=node_type_id,
                )
            )
        self.log.info(f"{MSG}: Wrote {override_path.relative_to(self.project_path)}")

    def write_databricks_run_script(self):
        MSG = "Creating Databricks run script"
        script_path = (
            self.project_path / "src" / self.package_name / "databricks_run.py"
        )
        toml_path = self.project_path / "pyproject.toml"
        shutil.copy(TEMPLATES / "databricks_run.py", script_path)
        self.log.info(f"{MSG}: Wrote {script_path.relative_to(self.project_path)}")

        with open(toml_path) as f:
            toml = tomlkit.load(f)

        scripts = toml.get("project", {}).get("scripts", {})
        if "databricks_run" not in scripts:
            scripts["databricks_run"] = f"{self.package_name}.databricks_run:main"
            toml["project"]["scripts"] = scripts

        self.log.info(
            f"{MSG}: Added script to {toml_path.relative_to(self.project_path)}"
        )
        with open(toml_path, "w") as f:
            tomlkit.dump(toml, f)

    def substitute_catalog_paths(self):
        MSG = "Substituting DBFS paths"
        conf_dir = self.metadata.project_path / "conf"
        envs = [d for d in conf_dir.iterdir() if d.is_dir()]
        regex = r"(.*/dbfs/FileStore/)(.*)(/data.*)"
        for env in envs:
            path = conf_dir / env / "catalog.yml"
            self.log.info(f"{MSG}: Checking {path.relative_to(self.project_path)}")

            if not path.exists():
                self.log.warning(
                    f"{MSG}: {path.relative_to(self.project_path)} does not exist."
                )
                continue

            with open(path) as f:
                content = f.readlines()

            new_content = self._parse_content(regex, path, content)

            with open(path, "w") as f:
                f.writelines(new_content)

    def _parse_content(self, regex, path, content):
        new_content = []
        for line in content:
            new_line = re.sub(regex, f"\\g<1>{self.package_name}\\g<3>", line)
            if new_line != line:
                self.log.info(
                    f"{path.relative_to(self.project_path)}: "
                    f"Substituted: {line.strip()} -> {new_line.strip()}"
                )
            new_content.append(new_line)
        return new_content



from __future__ import annotations

from pathlib import Path

import pytest
import yaml

from kedro_databricks.init import NODE_TYPE_MAP, InitController
from kedro_databricks.utils import has_databricks_cli
from tests.utils import reset_init


def _write_dummy_catalog(catalog_path: Path):
    catalog = {
        "test": {
            "type": "pandas.CSVDataSet",
            "filepath": "/dbfs/FileStore/my-project/data/01_raw/test.csv",
        },
        "test2": {
            "type": "pandas.CSVDataSet",
            "filepath": "/dbfs/FileStore/my-project/data/01_raw/test2.csv",
        },
    }
    with open(catalog_path, "w") as f:
        f.write(yaml.dump(catalog))


def test_substitute_catalog_paths(metadata):
    controller = InitController(metadata)
    catalog_path = controller.project_path / "conf" / "base" / "catalog.yml"
    _write_dummy_catalog(catalog_path)
    controller.substitute_catalog_paths()
    with open(catalog_path) as f:
        new_catalog = f.read()
    assert "/dbfs/FileStore/fake_project/data" in new_catalog, new_catalog


def test_write_override_template(metadata):
    controller = InitController(metadata)
    default_key = "default"
    provider = "azure"

    controller.write_kedro_databricks_config(default_key, provider)
    override_path = Path(metadata.project_path) / "conf" / "base" / "databricks.yml"
    assert override_path.exists(), "Override template not written"

    with open(override_path) as f:
        override = yaml.safe_load(f)
    assert (
        override.get(default_key) is not None
    ), f"Override template not written: {override}"

    job_cluster = override.get(default_key, {}).get("job_clusters", []).pop()
    assert (
        job_cluster.get("job_cluster_key") == default_key
    ), f"job_cluster_key is wrong: {override}"

    node_type_id = job_cluster.get("new_cluster", {}).get("node_type_id")
    assert node_type_id == NODE_TYPE_MAP[provider], f"node_type_id is wrong: {override}"

    try:
        controller.write_kedro_databricks_config(default_key, provider)
    except Exception:
        pytest.fail("If an override file already exists, it should not be overwritten.")


def test_write_databricks_run_script(metadata):
    controller = InitController(metadata)
    controller.write_databricks_run_script()
    run_script_path = (
        Path(controller.project_path)
        / "src"
        / controller.package_name
        / "databricks_run.py"
    )
    assert run_script_path.exists(), "Databricks run script not written"


def test_bundle_init(metadata):
    controller = InitController(metadata)
    if not has_databricks_cli():
        with pytest.raises(Exception):
            controller.bundle_init()
    else:
        reset_init(metadata)
        controller.bundle_init()
        bundle_path = Path(metadata.project_path) / "databricks.yml"
        if not bundle_path.exists():
            files = [
                f.relative_to(metadata.project_path).as_posix()
                for f in metadata.project_path.iterdir()
            ]
            pytest.fail(
                "Bundle file not written - found files:\n\t{}".format(
                    "\n\t".join(files)
                )
            )

        bundle = yaml.load(bundle_path.read_text(), Loader=yaml.FullLoader)
        assert (
            bundle is not None
        ), f"Bundle template failed to load - {bundle_path.read_text()}"
        assert bundle.get("bundle", {}).get("name") == metadata.package_name, bundle

        try:
            controller.bundle_init()
        except Exception:
            pytest.fail(
                "If a bundle file already exists, it should not be overwritten."
            )



# This file is used to run Kedro pipelines on Databricks.
# It is automatically generated by the Kedro-Databricks plugin.
# Do not modify this file directly.

import argparse
import logging

from kedro.framework.project import configure_project
from kedro.framework.session import KedroSession


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--env", dest="env", type=str)
    parser.add_argument("--conf-source", dest="conf_source", type=str)
    parser.add_argument("--package-name", dest="package_name", type=str)
    parser.add_argument("--nodes", dest="nodes", type=str)

    args = parser.parse_args()
    env = args.env
    conf_source = args.conf_source
    package_name = args.package_name
    nodes = [node.strip() for node in args.nodes.split(",")]

    # https://kb.databricks.com/notebooks/cmd-c-on-object-id-p0.html
    logging.getLogger("py4j.java_gateway").setLevel(logging.ERROR)
    logging.getLogger("py4j.py4j.clientserver").setLevel(logging.ERROR)

    configure_project(package_name)
    with KedroSession.create(env=env, conf_source=conf_source) as session:
        if len(nodes) > 0:
            session.run(node_names=nodes)
        else:
            session.run()


if __name__ == "__main__":
    main()



from kedro_databricks.plugin import commands
from tests.utils import reset_init


def test_deploy(cli_runner, metadata):
    """Test the `deploy` command"""
    reset_init(metadata)
    deploy_fail = ["databricks", "deploy"]
    result = cli_runner.invoke(commands, deploy_fail, obj=metadata)
    assert result.exit_code == 1, (result.exit_code, result.stdout)

    init_cmd = ["databricks", "init"]
    result = cli_runner.invoke(commands, init_cmd, obj=metadata)
    override_path = metadata.project_path / "conf" / "base" / "databricks.yml"
    assert result.exit_code == 0, (result.exit_code, result.stdout)
    assert metadata.project_path.exists(), "Project path not created"
    assert metadata.project_path.is_dir(), "Project path is not a directory"
    assert override_path.exists(), "Override file not created"

    deploy_cmd = ["databricks", "deploy", "--bundle", "--debug"]
    result = cli_runner.invoke(commands, deploy_cmd, obj=metadata)
    assert result.exit_code == 0, (result.exit_code, result.stdout)


def test_deploy_with_conf(cli_runner, metadata):
    """Test the `deploy` command"""
    reset_init(metadata)
    deploy_fail = ["databricks", "deploy"]
    result = cli_runner.invoke(commands, deploy_fail, obj=metadata)
    assert result.exit_code == 1, (result.exit_code, result.stdout)

    init_cmd = ["databricks", "init"]
    result = cli_runner.invoke(commands, init_cmd, obj=metadata)
    override_path = (
        metadata.project_path / "conf" / "sub_pipeline" / "base" / "databricks.yml"
    )
    override_path.parent.mkdir(parents=True, exist_ok=True)
    override_path.write_text(
        """
        default:
            job_clusters:
                - job_cluster_key: default
                new_cluster:
                    spark_version: 14.3.x-scala2.12
                    node_type_id: Standard_DS3_v2
                    num_workers: 1
                    spark_env_vars:
                        KEDRO_LOGGING_CONFIG: "/dbfs/FileStore/develop_eggs/conf/logging.yml"
            tasks:
                - task_key: default
                job_cluster_key: default
        """
    )
    assert result.exit_code == 0, (result.exit_code, result.stdout)
    assert metadata.project_path.exists(), "Project path not created"
    assert metadata.project_path.is_dir(), "Project path is not a directory"

    deploy_cmd = [
        "databricks",
        "deploy",
        "--bundle",
        "--debug",
        "--conf=conf/sub_pipeline",
    ]
    result = cli_runner.invoke(commands, deploy_cmd, obj=metadata)
    assert result.exit_code == 0, (result.exit_code, result.stdout)
